@@ -64,10 +64,22 @@ def mk_mall_of_dill_stores(store_names=Iterable[StoreName], rootdir=None):
     return {name: DillFiles(root) for name, root in name_and_rootdir()}
 
 
-# TODO: store_on_output: use i2.wrapper and possibly extend i2.wrapper to facilitate
 from functools import wraps
+from i2 import call_forgivingly
 
 
+def _validate_function_keyword_only_params(func, allowed_params: Iterable, obj_name):
+    if func is not None:
+        if not callable(func):
+            raise TypeError(f'{obj_name} should be callable: {func}')
+        sig = Sig(func)
+        if not all(kind == Parameter.KEYWORD_ONLY for kind in sig.kinds.values()):
+            raise TypeError(f'All params of {obj_name} must be keyword-only. {sig}')
+        if not set(sig.names).issubset(allowed_params):
+            raise TypeError(f'All params of {obj_name} must be in {allowed_params}')
+
+
+# TODO: store_on_output not pickalable: extend i2.wrapper to be able to solve with it
 @double_up_as_factory
 def store_on_output(
     func=None,
@@ -75,19 +87,106 @@ def store_on_output(
     store=None,
     save_name_param='save_name',
     add_store_to_func_attr='output_store',
+    empty_name_callback=None,
+    auto_namer=None,
+    output_trans=None,
 ):
     """Wrap func so it will have an extra save_name_param that can be used to
     indicate whether to save the output of the function call to that key, in
-    that store
+    that store.
+
+    The store can be specified, but an empty dict will be made for it by default.
+
+    By default, a pointer to this store is added to the wrapped function so it can be
+    used for inspection etc.
+
+    :param func: Function that we're wrapping
+    :param store: Mapping (e.g. dict) where we'd like the output to be stored in
+    :param save_name_param: Name of the extra param that will be added to the
+        function's interface, where the user will be able to specify the key where
+        they want to save the output.
+    :param add_store_to_func_attr: If not None, the wrapped function will have an
+        attribute of that name where the store will be stored
+    :param empty_name_callback: If not None, will be called if the user doesn't specify
+        a save name. Intended use; raising errors to force the user to enter name.
+    :param auto_namer: If not None, should be a keyword-only ``(*, arguments, output)``
+        callback that will be called (forgivingly). This callback will
+        be called when the user doesn't specify a name (or an empty name). It's output
+        will be used as the ``save_name``.
+    :param output_trans: If not None, should be a keyword-only
+        ``(*, save_name, output, arguments)`` function that will be called, returning
+        it's result instead of the ``output``.
+    :return: A wrapped function that has an extra param (default is ``save_name``) that
+        allowing the user to specify a key under which to save the output.
+
+    A simple example:
+
+    >>> def foo(a, b: int = 2):
+    ...     return a + b
+    >>> f = store_on_output(foo)
+    >>> f(2, 3, save_name='take')
+    5
+    >>> f.output_store
+    {'take': 5}
+
+    An example with auto_namer. Note that an auto_namer can have no parameters (e.g.
+    return a stringified timestamp for the auto-name), but if it does have any
+    parameters, these parameters must be keyword-only and the only names allowed are:
+
+    - ``save_name``, which will contain the value of the name the user entered
+
+    - ``output``, which will contain the return value of the function call
+
+    - ``arguments``, which will contain a dict of all the arguments of the function call
+
+    In the following, we will make an ``output_trans`` that just returns the
+    ``save_name``. This is useful when you want to make a pipeline and tell the next
+    function where to find the output of a previous function.
+
+    >>> def return_key(*, save_name):
+    ...     return save_name
+    >>> g = store_on_output(foo, output_trans=return_key)
+    >>> g(100, 23, save_name='test')
+    'test'
+    >>> g.output_store
+    {'test': 123}
+    >>> g.output_store[g(40, 2, save_name='here')]
+    42
+
+    An example involving more params:
+
+    >>> my_store = {'all': 'mine'}
+    >>> @store_on_output(
+    ...     store=my_store,
+    ...     save_name_param='save_as',
+    ...     add_store_to_func_attr=None,
+    ...     auto_namer=lambda *, arguments: '_'.join(map(str, arguments.values()))
+    ... )
+    ... def bar(a, b: int = 2):
+    ...     return a + b
+    >>> bar(2, 3, save_as='test')
+    5
+    >>> my_store
+    {'all': 'mine', 'test': 5}
+    >>> bar(7)
+    9
+    >>> my_store
+    {'all': 'mine', 'test': 5, '7_2': 9}
 
-    :param func:
-    :param store:
-    :param save_name_param: Name of the extra param
-    :return:
     """
     save_name_param_obj = Parameter(
         name=save_name_param, kind=Parameter.KEYWORD_ONLY, default='', annotation=str,
     )
+    _validate_function_keyword_only_params(
+        auto_namer, ['output', 'arguments'], obj_name='auto_namer'
+    )
+    _validate_function_keyword_only_params(
+        output_trans, ['save_name', 'output', 'arguments'], obj_name='output_trans'
+    )
+    if output_trans:
+        assert callable(output_trans) and set(Sig(output_trans).names).issubset(
+            ['save_name', 'output', 'arguments']
+        )
     sig = Sig(func) + [save_name_param_obj]
 
     if store is None:
@@ -96,19 +195,26 @@ def store_on_output(
     @sig
     @wraps(func)
     def _func(*args, **kwargs):
-        kwargs = sig.kwargs_from_args_and_kwargs(args, kwargs, apply_defaults=True)
-        save_name = kwargs.pop(save_name_param)
-        args, kwargs = sig.args_and_kwargs_from_kwargs(kwargs)
+        arguments = sig.kwargs_from_args_and_kwargs(args, kwargs, apply_defaults=True)
+        save_name = arguments.pop(save_name_param)
+        if not save_name and empty_name_callback:
+            empty_name_callback()
+        args, kwargs = sig.args_and_kwargs_from_kwargs(arguments)
         output = func(*args, **kwargs)
+        if not save_name and auto_namer:
+            save_name = call_forgivingly(auto_namer, arguments=arguments, output=output)
         if save_name:
             store[save_name] = output
-        return output
+        if not output_trans:
+            return output
+        else:
+            return call_forgivingly(
+                output_trans, save_name=save_name, output=output, arguments=arguments
+            )
 
     if isinstance(add_store_to_func_attr, str):
         setattr(_func, add_store_to_func_attr, store)
 
-    # _func.output_store = store  # redundant with above. Remove
-
     return _func
 
 
